/*
LCD drive
  Pin assignment
  1:VDD:P:5V(2.7V TO 5V)
  2:RA5:O:LCD RS
  3:RA4:I:AD Converter
  4:RA3/MCLR:I:None
  5:RC5:O:P PWM pulse out
  6:RC4:0:LCD E
  7:RC3:O:LCD D7

  14:VSS:Power GND
  13:RA0:I:AD Converter
  12:RA1:I:AD Converter
  11:RA2:I:AD Converter
  10:RC0:O:LCD D4
   9:RC1:0:LCD D5
   8:RC2:O:LCD D6
*/

#include <xc.h>
//#include <htc.h>
#include  "lcdwo.h"

//-LCD kind settings **1.20 -------------------------------
#ifdef  _LCD1602
#define  LCD_LineLength1  0x10  //LCD 16x2 **1.20
#endif
#ifdef  _LCD0802
#define  LCD_LineLength1  0x08  //LCD 8x2 **1.20
#endif
#ifdef  _LCD2402
#define  LCD_LineLength1  0x18  //LCD 24x2 **1.20
#endif
//--------------------------------------------------------

#define  LCD_LineLength2  (LCD_LineLength1 + LCD_LineNumber2)  // **1.20

static bit LCD_BF  @ ((unsigned)&PORTC*8+3);  // Busy Flag (RC3)
//static bit LCD_RW  @ ((unsigned)&PORTA*8+4);  // Read/write (RA4)
static bit LCD_RS  @ ((unsigned)&PORTA*8+5);  // Register select (RA5)
static bit LCD_EN  @ ((unsigned)&PORTC*8+4);  // Enable (RC4)
#define  LCD_STROBE()    NOP();LCD_EN=1;NOP();NOP();LCD_EN=0;   // spec. min 220ns
    // 前に250ns時間稼ぎし、Highで250ns保持

/*
// LCD busy wait
//  LCD_RSを事前に設定しておく必要有り
void lcd_busy_wait(void){

//__delay_us(100);      // for test
//return;               // for test

//TRISC =  0b11001111;  // RC3-0 Input
  TRISC =  (TRISC & 0b11110000) | 0b00001111;  // RC3-0 Input
  LCD_RW = 1;           // Read mode
  while(1){
    LCD_STROBE();
    if(!LCD_BF){
      TRISC =  (TRISC & 0b11110000) | 0b00000000;  // RC3-0 Output
//    TRISC =  0b11000000;  // RC5-0 All Output
      LCD_STROBE();         // 下位4bits読み飛ばし
      LCD_RW = 0;           // write mode
      break;
    }
    LCD_STROBE();           // 下位4bits読み飛ばし
  }
}
*/

// Write a 4bit control word to the LCD in 4 bit mode
//  LCD_RS=0を事前に設定しておく必要有り
void lcd_data_byteset(unsigned char c){
  PORTC = (PORTC & 0b11110000) |  (c & 0b00001111);    //Only low 4bit DATA
  LCD_STROBE();
}

// Write a control byte to the LCD in 4 bit mode
//  LCD_RS=0を事前に設定しておく必要有り
void lcd_write(unsigned char c){
  PORTC = (PORTC & 0b11110000) | ( (c >>4 ) & 0b00001111 ) ;  //Hi SIDE DATA
  LCD_STROBE();
  PORTC = (PORTC & 0b11110000) |  (c & 0b00001111 );    //low SIDE DATA
  LCD_STROBE();
//  lcd_busy_wait();
  __delay_us(40);
}

// Clear and home the LCD
void lcd_clear(void)
{
  LCD_RS = 0;  // control mode
  lcd_write(LCD_Clear);
  lcd_loccnt=0;
  __delay_ms(2); //PICC Pro
}

//Go to the specified position
void lcd_goto(unsigned char pos)
{
  LCD_RS = 0;  // control mode
  lcd_write(LCD_HomePos + pos);
  lcd_loccnt=pos;
}

// Cursor Off the LCD
void lcd_cursoroff(void)
{
  LCD_RS = 0;  // control mode
  lcd_write(LCD_CursorOff);
}

/*
// Cursor On the LCD
void lcd_cursoron(void)
{
  LCD_RS = 0;  // control mode
  lcd_write(LCD_CursorOn);
}
*/
//write a string of chars to the LCD
void lcd_puts(const char * s)
{
  while(*s)
    lcd_putch(*s++);
}

// Write a character to the LCD in 4 bit mode
void lcd_data_set(unsigned char c){
  LCD_RS = 1;  // character mode
  PORTC = (PORTC & 0b11110000) | ( (c >>4 ) & 0b00001111 ) ;  //Hi SIDE DATA
  LCD_STROBE();
  PORTC = (PORTC & 0b11110000) |  (c & 0b00001111 );    //low SIDE DATA
  LCD_STROBE();
  LCD_RS = 0;  // control mode
//  lcd_busy_wait();
  __delay_us(40);
}

//write one character to the LCD and auto increment location counter*/
void lcd_putch(unsigned char c){
  lcd_data_set(c);

  lcd_loccnt++;
  if (lcd_loccnt == LCD_LineLength1){
    lcd_loccnt = LCD_LineNumber2;
    lcd_goto(lcd_loccnt);
  }
  if (lcd_loccnt == LCD_LineLength2){
    lcd_loccnt = LCD_LineNumber1;
    lcd_goto(lcd_loccnt);
  }
}

//write to CG RAM 
void lcd_cgwrite(unsigned char c)
{
  lcd_data_set(c);
}

//initialise the LCD - put into 4 bit mode
void lcd_init(void)
{
//  LCD_RW = 0;          // write mode
  LCD_RS = 0;          // write control bytes
  __delay_ms(15);      // power on delay (PICC Pro)
  lcd_data_byteset(0x03);  // 8bit mode set
  __delay_ms(5);       // wait 5ms
  LCD_STROBE();        // 8bit mode set again
  __delay_us(100);     // wait 100us
  LCD_STROBE();        // 8bit mode set again
  __delay_ms(5);       // wait 5ms
  lcd_data_byteset(0x02) ;// set 4 bit mode
  __delay_us(40);     // wait 40us
  lcd_write(0x28);     // 4 bit mode, 1/16 duty, 5x8 font
  lcd_write(0x08);     // display off
  lcd_write(0x0F);     // display on, blink curson on
  lcd_write(0x06);     // entry mode
  lcd_loccnt=0;
}

/*
//Newline processing
void lcd_newline(unsigned char mode){
//  unsigned char temp;
  switch (mode){
    case 1://the following are deleted
      if (lcd_loccnt <= LCD_LineLength1){
        lcd_space_full(LCD_LineLength1-lcd_loccnt);
        return;
      }else if (lcd_loccnt <= LCD_LineLength2) {
        lcd_space_full(LCD_LineLength2-lcd_loccnt);
        return;
      }else return;

    default://toggle the line
      if (lcd_loccnt <= LCD_LineLength1) lcd_goto(LCD_LineNumber2);
      else if (lcd_loccnt <= LCD_LineLength2) lcd_goto(LCD_LineNumber1);
  }
}
*/
/*
void lcd_space_full(unsigned char count){//spacing
  while(count--){
    lcd_putch(' ');//space character write
  };
}
*/