/**************************************************************
 鉛蓄電池充電制御プログラム(AD変換main内無限ループ方式)
  From BaseRoad 51-55V  To 12V/24Vバッテリ(定電流/定電圧モード制御)充電
   for PIC16F1823(32MHz clock) PWM 40kHz
 ***************************************************************
  author Toshihiko HORIMA
  Tools:
  MPLAB X IDE v1.60
  MPLAB XC8 compiler ver. 1.12
---------------------------------------------------------------
Date  :Time  Version  Build  Note
20130324:0929  0.01  ---  modified the program
* ---------------------------------------------------------------
  Pin assignment
  1:VDD:P:5V(2.7V TO 5V)
  2:RA5:O:LCD RS
  3:RA4:I:AD Converter(Battery Charge Current)
  4:RA3/MCLR:I:14V/32V switch(Lo:14V/Hi:32V)
  5:RC5:O:P PWM pulse out
  6:RC4:0:LCD E
  7:RC3:O:LCD D7

  14:VSS:Power GND
  13:RA0:I:AD Converter(Baseload Voltage)
  12:RA1:I:AD Converter(Output Voltage)
  11:RA2:I:AD Converter(Battery Charge Voltage)
  10:RC0:O:LCD D4
   9:RC1:0:LCD D5
   8:RC2:O:LCD D6
---------------------------------------------------------------

**************************************************************/
#include <xc.h>
#include <stdio.h>
#include "lcdwo.h"

#define MHz 000000
#define _XTAL_FREQ 32MHz
//#define _25V-IN
#define _50V-IN
#define _14V-OUT
//#define _32V-OUT
//#define _20A
#define _5A

// 電圧・電流制御のための値設定
// 鉛蓄電池への充電のための設定
// 電流測定:ACS712電流センサ(20A)100mV/1A
// 入力電圧測定:分圧 64V:4.096V
// 出力電圧測定　(20V):分圧 20V:4.096V
// 　　　　　　　(42V):分圧 42V:4.096V

// 入力電圧値
#ifdef _25V-IN
#define BASEHIGH    0x210   // 33.0V(2.115V)   (2115)/4=529
#define BASELOW     0x140   // 20.0V(1.282V)   (1282)/4=320
#endif
#ifdef _50V-IN
#define BASEHIGH    0x3FC   // 63.8V(4.080V)   (4080)/4=1020
//#define BASELOW     0x32F   // 51.0V(3.262V)   (3262)/4=815
#ifdef _32V-OUT  // 抵抗値の誤差で14Vボードが優先されるようにするため(今は逆)
#define BASELOW     0x2B3   // 44.0V(2.814V)   (2814)/4=704  BASEがシャープ直列の場合
//#define BASELOW     0x2C0   // 44.0V(2.814V)   (2814)/4=704  BASEがシャープ直列の場合
#endif
#ifdef _14V-OUT
//#define BASELOW     0x2B7   // 43.5V(2.782V)   (2782)/4=696  BASEがシャープ直列の場合
#define BASELOW     0x337   // 51.5V(3.294V)   (3294)/4=823  BASEの場合
#endif
#endif
// 出力電圧値
// ACDelcoM31MF直列の場合(ターゲット電圧32V)
#ifdef _32V-OUT
#define BATTVOUTHIGH   0x3B7   // 35.0V(3.500V)   (3500)/4=951
#define BATTVOUTTGTCI  0x365   // 32.0V(3.200V)   (3200)/4=870
#define BATTVOUTTGTCV  0x2E3   // 27.2V(2.720V)   (2720)/4=739
#define BATTVOUTLOW    0x2B7   // 25.6V(2.560V)   (2560)/4=696
#endif
// 日立HXG 60B24Lの場合(ターゲット電圧14V)
#ifdef _14V-OUT
#define BATTVOUTHIGH   0x1B2   // 16.0V(1.600V)   435
//#define BATTVOUTTGTCI  0x18A   // 14.5V(1.450V)   394
#define BATTVOUTTGTCI  0x17C   // 14.0V(1.400V)   380
#define BATTVOUTTGTCV  0x171   // 13.6V(1.360V)   370
#define BATTVOUTLOW    0x15B   // 12.8V(1.280V)   348
#endif
// 出力電流値
#ifdef _20A
#define BATTIOUTLIMIT  0xC7   // 8A(ACS712:0.8V) (800)/4=200
#define BATTIOUTTARGET 0x7D   // 5A(ACS712:0.5V) (500)/4=125
#define BATTIOUTSWITCH 0x19   // 1A(ACS712:0.1V) (100)/4=25
#define BATTIOUTMINI   0x03   // 0.12A(ACS712:0.012V) (12)/4=3
#endif
#ifdef _5A
#define BATTIOUTLIMIT 0x10E   // 6A(ACS712:0.8V)
#define BATTIOUTTARGET 0xE0   // 5A(ACS712:0.5V)
#define BATTIOUTSWITCH 0x2D   // 1A(ACS712:0.1V)
#define BATTIOUTMINI   0x05   // 0.12A(ACS712:0.012V) (12)/4=3
#endif
// チャージポンプ再開判定用出力電圧値
#define VOUTHIGH        0x285  // 40.0V(2.584V)   (2584)/4=646
#define VOUTMID         0xA1   // 10.0V(0.646V)    (646)/4=161
#define CP_START        0x30   //  3.0V(0.193V)    (193)/4=48

// AD channel
#define V_BASE          0  // 出力電圧用:ADチャネル0
#define V_OUT           1  // 出力電圧(ダイオード前)用:ADチャネル1
#define V_CHARGE        2  // 出力電圧用:ADチャネル2
#define I_CHARGE        3  // 出力電流用:ADチャネル3

/* ベース電圧状態コード
     ↑   BASE_HIGHER    V2
  BASEHIGH
     ｜   BASE_NORMAL    V1
  BASELOW
     ↓   BASE_LOW       V0
*/
#define BASE_HIGHER        2 // 入力電圧限界以上(緊急停止)
#define BASE_NORMAL        1 // 入力電圧正常範囲
#define BASE_LOW           0 // 入力電圧下限アンダー(入力電圧低下休止)

/* 充電電圧状態コード
     ↑   BATTVOUT_HIGHER  BATTVOUT_HIGH_EST V4
  BATTVOUTHIGH
     ｜   BATTVOUT_HIGH     V3
  BATTVOUTTGTCI
     ｜   BATTVOUT_MID      V2
  BATTVOUTTGTCV
     ｜   BATTVOUT_LOW      V1
  BATTVOUTLOW
     ↓   BATTVOUT_LOWER    V0
*/
#define BATTVOUT_HIGHER       5 // 充電停止電圧以上(充電停止)
#define BATTVOUT_HIGH_EST     4 // 充電停止電圧以上(評価中)
#define BATTVOUT_HIGH         3 // 定電流ターゲット電圧値以上
#define BATTVOUT_MID          2 // 定電流ターゲット電圧値以下
#define BATTVOUT_LOW          1 // 定電圧ターゲット電圧値以下
#define BATTVOUT_LOWER        0 // 出力電圧下限アンダー(電池放電停止)

/* 充電電流状態コード
     ↑   BATTIOUT_HIGHER I6  BATTIOUT_HIGH_EST I5
  BATTIOUTLIMIT
     ｜   BATTIOUT_HIGH     I4
  BATTIOUTTARGET
     ｜   BATTIOUT_MID      I3
  BATTIOUTSWITCH
     ｜   BATTIOUT_LOW      I2
  BATTIOUTMINI
     ↓   BATTIOUT_SLEEP    I1
  BATTIOUTZERO ±0
     ↓   BATTIOUT_REV 逆流 I0
*/
#define BATTIOUT_HIGHER       6 // 限界電流以上(緊急停止)
#define BATTIOUT_HIGH_EST     5 // 限界電流以上(評価中)
#define BATTIOUT_HIGH         4 // 定電流ターゲット以上
#define BATTIOUT_MID          3 // 定電流ターゲット以下
#define BATTIOUT_LOW          2 // 定電圧切替えターゲット以下
#define BATTIOUT_SLEEP        1 // 定電圧切替えターゲット以下
#define BATTIOUT_REV          0 // 電流逆流

/* 出力電圧状態コード
     ↑ VOUT_HIGH    V2
 VOUTHIGH
     | VOUT_MID      V1
  CP_START
     ↓ VOUT_STANDBY V0
*/
#define VOUT_HIGH         2 // 出力電圧限界以上
#define VOUT_MID          1 // 出力電圧正常範囲
#define VOUT_STANDBY      0 // 出力電圧チャージポンプ可能に低下

// 状態コード(mode_statusに設定される値)
#define STOP_CHARGE           0x00 // 入力電圧低下により休止中
//#define CAPACITY_CHARGE       0x01 // 出力コンデンサ充電
#define CONST_CURRENT         0x02 // 定電流急速充電モード(出力14or32V/5Amax)
#define CONST_VOLTAGE         0x03 // 定電圧トリクル充電モード(出力13.6or27.2V)
#define FULL_CHARGED          0x04 // 満充電(出力12.8V)
#define RECOVERY_WAIT         0x07 // PWM停止後に出力電圧が低下するのを待っている
#define ERROR_CODE            0x80 // エラーコード
#define ERR_BASEHIGH_OVER     0b10000001 // ベース電圧限界オーバー(緊急停止) ↓以降異常
#define ERR_VOUTHIGH_OVER     0b10000010 // 出力電圧限界オーバー(緊急停止)
#define ERR_BATTIOUTHIGH_OVER 0b10000100 // 充電電流限界オーバー(緊急停止)
#define ERR_BATTVOUTHIGH_OVER 0b10001000 // 充電電圧限界オーバー(緊急停止)
#define ERR_INVALID_INT       0b11000000 // 割込み異常

#define PWM_PERIOD      199 // PWMの周期幅(40kHz)
#define AVE_NUM           8 // 測定値の平均をとる回数
#define AVE_NUM2          7 // 測定値の平均をとる回数
#define DISPLAY_TIME     61 // 500/8.192=61 500ms間隔で表示

// CONFIG1ワードの設定
__CONFIG(
  FOSC_INTOSC & WDTE_OFF & PWRTE_ON & MCLRE_OFF
  & CP_OFF & CPD_OFF & BOREN_ON & CLKOUTEN_OFF
  & IESO_OFF & FCMEN_OFF
  );
// CONFIG2ワードの設定
__CONFIG(
  WRT_OFF & PLLEN_ON & STVREN_ON & LVP_OFF
//WRT_OFF & PLLEN_OFF & STVREN_ON & LVP_OFF
  & BORV_LO  //
  );

unsigned short int avechi,avechv,sumchi=0,sumchv=0,advalue[4];  // ADチャネルの測定値保存配列 0:入電圧,1:出電流,2:出電圧,3:出電圧
unsigned short int chv[AVE_NUM],chi[AVE_NUM];  // AD_NEM回分の測定値保存用
unsigned char hex_num[]={'0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f'};
unsigned char base_status,chv_status,chi_status,vout_status;  // 状態コード
short int valchi;         // 充電電流値
unsigned short int ioutzero;  // ioutzero初期値(2.5V)設定
unsigned short int cnt0=0;    // ディスプレイ表示タイミング用タイマ
short int duty=0;         // PWMパルス幅(0-199)
unsigned char mode_code, err_code, chi_cnt;  // 動作モードコード,エラーコード,電流超過時間カウンタ
unsigned char adcount=0, precnt;  // 平均値をとるためのカウンタ(0～7サイクル)

void delay_10ms(unsigned int temp){  //10ms multiple x times
  while(temp--){
    __delay_ms(10);
  }
}

// 左スペースパディング付き整数→文字列(16進4桁)変換
//   除算(/1000,/100など)はメモリを20%以上消費するため使えずシフト演算にした
//   呼び出し側で必要な桁数を得る場合は、(buf+1)などでアクセスすればよい
unsigned char* utoa_padding(unsigned short int num, unsigned char keta){
  static unsigned char buf[5];
  buf[keta--]='\0';
  do{
    buf[keta]=hex_num[num&0x000F];
    num/=16;
  }while(keta--);
  return buf;
}

// LCD1行目表示(モード、測定値モード、デューティー比)
void print_lcd_line1(void){
  lcd_goto(LCD_LineNumber1);
  lcd_puts("M");
  lcd_puts(utoa_padding(mode_code,2));
//  lcd_putch(hex_num[mode_code]);
  lcd_puts(" B");
  lcd_putch(hex_num[base_status]);
  lcd_puts(" I");
  lcd_putch(hex_num[chi_status]);
  lcd_puts(" V");
  lcd_putch(hex_num[chv_status]);
  lcd_puts(" D");
  lcd_puts(utoa_padding(duty,2));
}

// LCD2行目表示(測定値)
void print_lcd_line2(void){
// 表示用文字列(桁数+1)
//  lcd_goto(LCD_LineNumber2);
//  ベース電圧
  lcd_puts(utoa_padding(advalue[V_BASE],3));
  lcd_putch('V');
//  出力電圧
  lcd_puts(utoa_padding(advalue[V_OUT],3));
  lcd_putch('V');
//  充電電流
  lcd_puts(utoa_padding(valchi,3));
  lcd_putch('A');
//  充電電圧
  lcd_puts(utoa_padding(avechv,3));
  lcd_putch('V');
}

/*
//Write one character (correspond printf)
void putch(unsigned char c){
  lcd_putch(c);
  return;
}
*/

//  AD変換で入力電圧、出力電流、出力電圧を測定
void adconv(void){
  unsigned char adch=3; // ADチャネル(0:充電電圧,1:充電電流,2:ベース電圧,3:出力電圧)
  do{
    ADCON0 = (adch<<2)+1;    // 測定するADチャネルを設定
    __delay_us(8);           // 測定電圧Sample/Hold回路充電時間待ち(要計算)
    GIE = 0;           // 割込み禁止(AD変換中は割込み禁止)
    GO_nDONE = 1;      // AD変換開始
    while(GO_nDONE);         // AD変換完了を待つ(1823のバグのため使えず)
//    __delay_us(10);            // AD変換は開始後10.5～11TADの間に終了させる必要がある
    GO_nDONE = 0;      // AD変換終了
    GIE = 1;           // 割込み許可
    advalue[adch] = (ADRESH<<8)+(ADRESL); // 変換結果を読込み
  }while(adch--);
}

//  LCDディスプレイ表示
void display(void){
  lcd_goto(LCD_LineNumber1);
  print_lcd_line1();
  print_lcd_line2();
}

//  電圧電流基準値の設定
//  キャリブレーションAD変換で入力電圧、出力電流、出力電圧を測定
void calibration(void){
  unsigned char i;
  CCPR1L  = 0;          // PWM停止
  lcd_goto(LCD_LineNumber1);
  lcd_puts("Calibrating.....");  // キャリブレーション中表示
  // PWM停止し安定を待ち電流電圧測定
  delay_10ms(10);  // 100ms休止
  i=AVE_NUM2;
//sumchv=0;
//sumchi=0;
  do{
    adconv();
    chv[i]=advalue[V_CHARGE];
    chi[i]=advalue[I_CHARGE];
    sumchv=sumchv+chv[i];
    sumchi=sumchi+chi[i];
  }while(i--);
  avechv = sumchv/AVE_NUM;      // 充電(蓄電池)電圧の平均値
  ioutzero = sumchi/AVE_NUM;    // ゼロ点値を設定(5V電圧基準値)
  valchi=ioutzero;              // ゼロ点値を表示するため
// 測定結果表示
  lcd_goto(LCD_LineNumber1);
  lcd_puts("Calibrate Finish");  // キャリブレーション結果
  print_lcd_line2();
  delay_10ms(300);  // 3s表示
  CCPR1L  = duty;       // PWM再開
}

/* 割込みハンドラ
    割込みが発生するとGIEはリセットされ割込み不許可となってこのルーチンが開始
    GIE以外のPEIE、ADIEは変化しないため最初に設定したままでOK
    タイマー0割込みは8.192ms間隔で発生し、61回(約500ms)に1回ディスプレイ表示を行う
    電流電圧測定値に基づきモード診断、duty比調整を実行
    液晶ディスプレイは(DISPLAY_TIME*8.192)ms毎に表示
    電流電圧値測定はmainループで実施
*/
void interrupt handler( void ){

// 割込みの種類を判定
  if(TMR0IF)                    // TMR0割込みだったら
  {
    TMR0IF = 0;                 // 割込みフラグクリア

// 前処理
    err_code = 0;               // エラーコードクリア
    avechv = sumchv/AVE_NUM;    // 電圧測定値平均
    avechi = sumchi/AVE_NUM;    // 電流測定値平均
    if(avechi>ioutzero)         // 電流平均値が±0より大きければ
      valchi=avechi-ioutzero;   // オフセット補正
    else{                       // 5V電圧変動したので
      ioutzero=avechi;          // ±0点値を再設定
      valchi=0;
    }
// AD結果の緊急停止処理コード設定
// 出力電圧値チェック
    if(advalue[V_OUT]>VOUTHIGH){  // 出力電圧限界上限超え
      vout_status = VOUT_HIGH;
      mode_code = ERROR_CODE;
      err_code |= ERR_VOUTHIGH_OVER;}
  else if(advalue[V_OUT]>CP_START)    // 正常電圧範囲
    vout_status = VOUT_MID;
  else                                // チャージポンプ開始可能電圧
    vout_status = VOUT_STANDBY;
// ベース電圧値チェック
    if(advalue[V_BASE]>BASEHIGH){       // 緊急停止:入力電圧上限値オーバー
      base_status = BASE_HIGHER;
      mode_code = ERROR_CODE;
      err_code |= ERR_BASEHIGH_OVER;}
    else if(advalue[V_BASE]>BASELOW)    // 入力電圧正常範囲
      base_status = BASE_NORMAL;
    else                                // 入力電圧低下
      base_status = BASE_LOW;
// 充電電流値チェック
    if(valchi>BATTIOUTLIMIT){    // 緊急停止:出力電流上限値オーバー
      if(chi_cnt>10){
        chi_status = BATTIOUT_HIGHER;
        mode_code = ERROR_CODE;
        err_code |= ERR_BATTIOUTHIGH_OVER;}
      else {
        chi_cnt++;
        chi_status = BATTIOUT_HIGH;}
    }
    else {
      chi_cnt=0;
      if(valchi>BATTIOUTTARGET)
        chi_status = BATTIOUT_HIGH;
      else if(valchi>BATTIOUTSWITCH)
        chi_status = BATTIOUT_MID;
      else if(valchi>BATTIOUTMINI)
        chi_status = BATTIOUT_LOW;
      else if(valchi>0)
        chi_status = BATTIOUT_SLEEP;
      else
        chi_status = BATTIOUT_REV;
    }
// 充電電圧値チェック
    if(avechv>BATTVOUTHIGH){    // 緊急停止:出力電圧上限値オーバー
      chv_status = BATTVOUT_HIGHER;
      mode_code = ERROR_CODE;
      err_code |= ERR_BATTVOUTHIGH_OVER;}
    else{
      if(avechv>BATTVOUTTGTCI)      // 定電流ターゲット電圧超え
        chv_status = BATTVOUT_HIGH;
      else if(avechv>BATTVOUTTGTCV) // 定電圧ターゲット電圧超え
        chv_status = BATTVOUT_MID;
      else if(avechv>BATTVOUTLOW)   // 定電圧ターゲット電圧以下
        chv_status = BATTVOUT_LOW;
      else                          // 放電停止電圧以下
        chv_status = BATTVOUT_LOWER;
    }
  }      // 割込みがTMR0の場合の処理エンド
  else{  // 割込みがTMR0でなかった場合
    mode_code = ERROR_CODE;
    err_code = ERR_INVALID_INT;  // 想定外の割込み発生
  }

// 緊急停止の場合はPWMクロックを停めてERR状況を表示し5秒停止し表示カウンタクリア
  if(mode_code>=ERROR_CODE){
    CCPR1L = 0;                 // TMR2(PWM)緊急停止
    duty = 0;                   // dutyクリア
    mode_code = err_code;       // モードコードにエラーコードを設定
    display();                  // LCD表示
    delay_10ms(500);            // 5s待ち
    chi_cnt=0;                  // 電流超過時間カウンタクリア
    cnt0=0;                     // ディスプレイ表示カウンタクリア
    mode_code = RECOVERY_WAIT;  // 復帰後は出力電圧低下待ちに
  }
  else{
/* 正常範囲の出力電流電圧制御処理
    鉛蓄電池充電制御
     入力電圧がBASE電圧下限(51V)を下回ったら充電停止
     停止で入力電圧が50Vを回復したらチャージポンプリカバリモードに
     チャージポンプリカバリで出力電圧が低下したら突入電流を制限するため
     コンデンサ充電モードに。充電完了したら定電流モードに
     定電流モードは出力電圧14.5V/32VになるまでMAX5Aで電流を流し続ける
     定電流モードで電圧が14.5V/32Vに達したら定電圧モードに移行
     定電圧モードはターゲット電圧13.6V/28Vを維持するようにPWM制御
     ターゲット電圧で電流が5A流れたら定電流モードに復帰
     定電圧モードでは13.6V/28Vで電流が徐々に少なくなりトリクル充電となる
*/
    if(base_status==BASE_LOW){ // ベース電圧が下限値以下になったら
      if(duty==0) mode_code=STOP_CHARGE;  // PWM停止中だったら停止
      else{
        duty--;                // PWM起動中だったらTMR2(PWM)幅狭める
        if(mode_code==CONST_CURRENT) mode_code=CONST_VOLTAGE;} // 定電流モードだったら定電圧モードに
    }
    else if((duty==PR2)&&(advalue[V_OUT]<CP_START))  // 出力電圧が低下したらチャージポンプ外れ
      mode_code = RECOVERY_WAIT;   // チャージポンプリカバリモードに
    else{                      // ベース電圧が正常範囲だったら
      switch(mode_code){       // 動作モードにより処理振り分け
      case CONST_CURRENT:      // 定電流モード
        if(chv_status>=BATTVOUT_HIGH) // 定電流終了電圧を上回った
          duty--;              // TMR2(PWM)幅狭める
        else if(chi_status>=BATTIOUT_HIGH) // ターゲット電流を上回った
          duty--;              // TMR2(PWM)幅狭める
        else if(chi_status<=BATTIOUT_LOW)  // 定電流圧切替え電流を下回った
          mode_code = CONST_VOLTAGE;
        else
          duty++;              // TMR2(PWM)幅拡げる
        break;
      case CONST_VOLTAGE:      // 定電圧モード
//        if(chv_status>=BATTVOUT_HIGH){ // 充電休止電圧を上回った
//          mode_code = FULL_CHARGED;  // 満充電完了
//          duty = 0;}                 // TMR2(PWM)停止
        if(chi_status>=BATTIOUT_MID)   // 定電圧ターゲット電流を上回った
          mode_code = CONST_CURRENT;
        else if(cnt0==0){   // 電圧変動はタイムラグがあるため0.5s毎にしか制御せず
          if(chv_status>=BATTVOUT_MID) // ターゲット電圧を上回った
            duty--;              // TMR2(PWM)幅狭める
          else
            duty++;}             //
        break;
      case FULL_CHARGED:       // 満充電モード
        if(chv_status<=BATTVOUT_LOW) // 定電圧ターゲット電圧を下回った
          mode_code = RECOVERY_WAIT; // 定電流モードに
        break;
      case STOP_CHARGE:        // 充電停止
        if(advalue[V_BASE]>BASELOW)  // 入力電圧が復帰した
          mode_code = RECOVERY_WAIT; // 定電流モードに
        break;
      case RECOVERY_WAIT:        // チャージポンプ開始待ち
        if(vout_status==VOUT_STANDBY){    // 開始
//          duty = 1;                    // dutyミニマムでコンデンサ充電
//          prevout = advalue[V_OUT];    // 充電開始初期出力電圧を設定
//          mode_code = CAPACITY_CHARGE;} // コンデンサ充電モードに
          mode_code = CONST_CURRENT;}  // 定電流モードに
        break;
/*      case CAPACITY_CHARGE:      // 出力コンデンサ充電中
        if((advalue[V_OUT]>VOUTMID)&&(prevout >= advalue[V_OUT]))  // 電圧値がサチったらコンデンサ満充電
          mode_code = CONST_CURRENT;   // 定電流モードに
        else                           // 充電継続
          prevout = advalue[V_OUT];    // 電圧値を保存
        break;*/
//    default:
//      break;
      }  // switch end
    }    // else end

//    if(duty>=PR2) duty=PR2;      // Duty値がPR2(199)を超えたらPR2に
    if(duty>=PR2){               // Duty値がPR2(199)を超えたら
      duty=0;                    // WPMを停止
      mode_code=RECOVERY_WAIT;}  // チャージポンプ外れとしリカバリ待ちモードに
    else if(duty<0) duty=0;      // Duty値が負になったら0にする
    CCPR1L = duty;               // 新しいPWM値を設定 CCPR1L/PR2[%]
//    if(mode_code==CONST_VOLTAGE) // 充電電圧の変化が追従するまでの待ち時間
//      __delay_ms(10);

// 500msに1回表示をリフレッシュ
    cnt0++;
    if(cnt0==DISPLAY_TIME){
      cnt0=0;                    // ディスプレイ表示カウンタクリア
      display();                 // LCDディスプレイ表示
    }
  } // 正常範囲の出力電流電圧制御処理エンド

  GIE = 1;                    // 割込み許可
}

//------------------main
main()
{

//  Port initalize
//           76543210
  OSCCON = 0b01110000;  // PLL On(configuration word2), 8MHz, IntOSC
  PORTA =  0x00;        // Port Initalize, all ports are Low output
  PORTC =  0x00;        // Port Initalize, all ports are Low output
  TRISA =  0b11011111;  // RA5:Out 4:In 3:In 2:In 1:In 0:In
  TRISC =  0b11000000;  // RC5-0 All Output
// Analog/Digital port setting
  ANSELA = 0b00010111;  // RA5:- 4:Ain 3:- 2:Ain 1:Ain 0:Ain
  ANSELC = 0x00;        // All port are use for digital port

// PWM初期化 (40.0kHz on RC5 @32MHz)
//  PWMサイクル:40.0kHz( (PR2+1)*4*Tosc*TMR2プリスケール )
//                       10kHz=(199+1)*4*(1/32MHz)*1)
//  分解能:PR2=199
//  duty 0～100% CCPR1L,CCP1CON<5:4> 0～800(0x320,0b001100100000)
//CCP1SEL = 0 ;         // 1822で必要。1823では設定不要
  CCP1CON = 0b00001100; //Single output,LSBset,PWM mode: P1A, P1C active-high; P1B, P1D active-high
//CCP1CON = 0b00001111; //Single output,LSBset,PWM mode: P1A, P1C active-low; P1B, P1D active-low
  T2CON   = 0b00000000; // POSTSCALE 1:1 ,TMR2 OFF , PRESCALE 1:1
  PR2     = PWM_PERIOD; // PWMの周期を設定(40kHzで設定)
  CCPR1H  = 0;

// TMR0初期化
//  定期的に電流、電圧値を測定し、モード設定するための割込みを発生させるために使用する
//  TMR0:(32MHz/4)*256*256=8.192ms 約8ms毎にタイマー割込みが発生する
//  割込みが発生すると入力電圧、出力電流、出力電圧を測定し、値によってモード設定、duty比調整
  OPTION_REG = 0b11000111; // PullUp:無効,clkソース:内部,PreScaler:有効,1:256

// AD Converter初期化
  FVRCON = 0b11000011;  // FVR enable, 4.096V for ADC
  ADCON0 = 0b00000001;  // b6-2:Ch0,b1:Go/Done b0:En/Da:最初はch0から
  ADCON1=  0b10100011;  // 読取値は右寄せ、A/D変換クロックはFOSC/32(32M-1us)、FVRをリファレンス

// メインループ前初期値設定
  mode_code = RECOVERY_WAIT; // チャージポンプリカバリ待ちモード
//  CCPR1L  = 0;  // キャリブレ中に設定するためここは不要
//  duty    = 0;  // 最初はデューティー比は0から(変数定義時に初期化)
  TMR0    = 0;  // モードチェック用(8ms)割込タイマー0を初期化
  TMR2    = 0;  // PWM用タイマー2を初期化

// LCD初期化
  delay_10ms(20);  //stabilize time
  lcd_init();      //lcd initalize
  lcd_clear();     //lcd clear
  lcd_cursoroff(); // Cursor Off

// LCD初期表示
// キャリブレーション
  calibration();

// メインループ前初期値設定
  TMR2ON  = 1;       // TMR2(PWM)スタート
  TMR0IE = 1;        // TMR0割込み有効
//  PEIE = 1;        // 周辺装置割込み有効(タイマー割込みには不要)
  GIE = 1;           // 割込み許可

/********************************
 メインループ
   電流電圧値測定を無限ループで実施
   充電電圧/充電電流はAVE_NUM回分の合計値を算出
   液晶ディスプレイ表示とモード診断、duty比調整はは割込みルーチンで実行
********************************/
  while(1){
//  AD変換で入力電圧、出力電流、出力電圧を測定
    adconv();
    chv[adcount]=advalue[V_CHARGE];
    chi[adcount]=advalue[I_CHARGE];
// AD_NUM回分の合計計算(今回分を加えAD_NUM回前を引く)
    precnt=adcount+1;
    if(precnt==AVE_NUM) precnt=0;
    sumchv=sumchv+chv[adcount]-chv[precnt];
    sumchi=sumchi+chi[adcount]-chi[precnt];
// 移動平均とるためのカウンタの設定(カウンタの初期化はキャリブレーションで実施)
    adcount++;
    if(adcount==AVE_NUM) adcount=0;
  }  // メインループエンド
}  // end of main()
